(*
(C) Copyright Andreas Viktor Hess, DTU, 2015-2018

All Rights Reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

- Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products
  derived from this software without specific prior written
  permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)

(*  Title:      Example_Keyserver.thy
    Author:     Andreas Viktor Hess, DTU
*)

theory Example_Keyserver
imports Stateful_Compositionality
begin

section {* Keyserver example setup *}
subsection {* Datatypes and functions setup *}
datatype ex_lbl = Label1 ("l1") | Label2 ("l2")

datatype ex_atom =
  Agent | Value | Attack | PrivFunSec
| Bot

datatype ex_fun =
  ring | valid | revoked | events | beginauth nat | endauth nat | pubkeys | seen
| invkey | pair | pair' | attack nat
| sign | crypt | update | pw
| encodingsecret | pubkey nat
| pubconst ex_atom nat

type_synonym ex_type = "(ex_fun, ex_atom) term_type"
type_synonym ex_var = "ex_type \<times> nat"

lemma ex_atom_UNIV:
  "(UNIV::ex_atom set) = {Agent, Value, Attack, PrivFunSec, Bot}"
by (auto intro: ex_atom.exhaust)

instance ex_atom::finite
by intro_classes (metis ex_atom_UNIV finite.emptyI finite.insertI)

instantiation ex_atom::enum
begin
definition "enum_ex_atom == [Agent, Value, Attack, PrivFunSec, Bot]"
definition "enum_all_ex_atom P ==
  list_all P [Agent, Value, Attack, PrivFunSec, Bot]"
definition "enum_ex_ex_atom P ==
  list_ex P [Agent, Value, Attack, PrivFunSec, Bot]"

instance
apply intro_classes
by (simp_all add: ex_atom_UNIV enum_ex_atom_def enum_all_ex_atom_def enum_ex_ex_atom_def)
end

lemma ex_lbl_UNIV:
  "(UNIV::ex_lbl set) = {Label1, Label2}"
by (auto intro: ex_lbl.exhaust)

instance ex_lbl::finite
by intro_classes (metis ex_lbl_UNIV finite.emptyI finite.insertI) 

instantiation ex_lbl::enum
begin
definition "enum_ex_lbl == [Label1, Label2]"
definition "enum_all_ex_lbl P == list_all P [Label1, Label2]"
definition "enum_ex_ex_lbl P == list_ex P [Label1, Label2]"

instance
apply intro_classes
by (simp_all add: ex_lbl_UNIV enum_ex_lbl_def enum_all_ex_lbl_def enum_ex_ex_lbl_def)
end


type_synonym ex_term = "(ex_fun, ex_var) term"
type_synonym ex_terms = "(ex_fun, ex_var) terms"

primrec arity::"ex_fun \<Rightarrow> nat" where
  "arity ring = 2"
| "arity valid = 3"
| "arity revoked = 3"
| "arity events = 1"
| "arity (beginauth _) = 3"
| "arity (endauth _) = 3"
| "arity pubkeys = 2"
| "arity seen = 2"
| "arity invkey = 2"
| "arity pair = 2"
| "arity pair' = 2"
| "arity (attack _) = 0"
| "arity sign = 2"
| "arity crypt = 2"
| "arity update = 4"
| "arity pw = 2"
| "arity (pubkey _) = 0"
| "arity encodingsecret = 0"
| "arity (pubconst _ _) = 0"

fun public::"ex_fun \<Rightarrow> bool" where
  "public f = (case f of
      (pubkey _) \<Rightarrow> False
    | encodingsecret \<Rightarrow> False
    | _ \<Rightarrow> True)"

fun Ana\<^sub>c\<^sub>r\<^sub>y\<^sub>p\<^sub>t::"ex_term list \<Rightarrow> (ex_terms \<times> ex_terms)" where
  "Ana\<^sub>c\<^sub>r\<^sub>y\<^sub>p\<^sub>t [k,m] = ({Fun invkey [Fun encodingsecret [], k]}, {m})"
| "Ana\<^sub>c\<^sub>r\<^sub>y\<^sub>p\<^sub>t _ = ({}, {})"

fun Ana\<^sub>s\<^sub>i\<^sub>g\<^sub>n::"ex_term list \<Rightarrow> (ex_terms \<times> ex_terms)" where
  "Ana\<^sub>s\<^sub>i\<^sub>g\<^sub>n [k,m] = ({}, {m})"
| "Ana\<^sub>s\<^sub>i\<^sub>g\<^sub>n _ = ({}, {})"

fun Ana::"ex_term \<Rightarrow> (ex_terms \<times> ex_terms)" where
  "Ana (Fun pair T) = ({}, set T)"
| "Ana (Fun pair' T) = ({}, set T)"
| "Ana (Fun sign T) = Ana\<^sub>s\<^sub>i\<^sub>g\<^sub>n T"
| "Ana (Fun crypt T) = Ana\<^sub>c\<^sub>r\<^sub>y\<^sub>p\<^sub>t T"
| "Ana _ = ({}, {})"


subsection {* Keyserver example: Locale interpretation *}
lemma assm1:
  "Ana t = (K,M) \<Longrightarrow> fv\<^sub>s\<^sub>e\<^sub>t K \<subseteq> fv t"
  "Ana t = (K,M) \<Longrightarrow> finite K"
  "Ana t = (K,M) \<Longrightarrow> (\<And>g S'. Fun g S' \<sqsubseteq> t \<Longrightarrow> length S' = arity g)
                \<Longrightarrow> k \<in> K \<Longrightarrow> Fun f T' \<sqsubseteq> k \<Longrightarrow> length T' = arity f"
  "Ana t = (K,M) \<Longrightarrow> K \<noteq> {} \<or> M \<noteq> {} \<Longrightarrow> Ana (t \<cdot> \<delta>) = (K \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>, M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>)"
by (rule Ana.cases[of "t"], auto elim!: Ana\<^sub>c\<^sub>r\<^sub>y\<^sub>p\<^sub>t.elims Ana\<^sub>s\<^sub>i\<^sub>g\<^sub>n.elims)+

lemma assm2: "Ana (Fun f T) = (K, M) \<Longrightarrow> M \<subseteq> set T"
by (rule Ana.cases[of "Fun f T"]) (auto elim!: Ana\<^sub>c\<^sub>r\<^sub>y\<^sub>p\<^sub>t.elims Ana\<^sub>s\<^sub>i\<^sub>g\<^sub>n.elims)

lemma assm5: "infinite {c. public c \<and> arity c = 0}"
proof -
  let ?\<tau> = "Bot"
  let ?T = "(range (pubconst ?\<tau>))::ex_fun set"
  have *:
       "\<And>x y::nat. x \<in> UNIV \<Longrightarrow> y \<in> UNIV \<Longrightarrow> (pubconst ?\<tau> x = pubconst ?\<tau> y) = (x = y)"
       "\<And>x::nat. x \<in> UNIV \<Longrightarrow> pubconst ?\<tau> x \<in> ?T"
       "\<And>y::ex_fun. y \<in> ?T \<Longrightarrow> \<exists>x \<in> UNIV. y = pubconst ?\<tau> x"
    by auto
  have "?T \<subseteq> {c. public c \<and> arity c = 0}" by auto
  moreover have "\<exists>f::nat \<Rightarrow> ex_fun. bij_betw f UNIV ?T"
    using bij_betwI'[OF *] by blast
  hence "infinite ?T" by (metis nat_not_finite bij_betw_finite)
  ultimately show ?thesis using infinite_super by blast
qed

lemma assm6: "0 < arity f \<Longrightarrow> public f" by (cases f) simp_all

interpretation intruder_model arity public Ana
apply standard
by (metis assm1(1), metis assm1(2), metis assm1(3), rule Ana.simps,
    metis assm2, metis assm1(4), metis assm5)

type_synonym ex_strand_step = "(ex_fun,ex_var) strand_step"
type_synonym ex_strand = "(ex_fun,ex_var) strand"


subsection {* Typing function *}
definition \<Gamma>\<^sub>v::"ex_var \<Rightarrow> ex_type" where
  "\<Gamma>\<^sub>v v = (if (\<forall>t \<in> subterms (fst v). case t of
                (TComp f T) \<Rightarrow> arity f > 0 \<and> arity f = length T
              | _ \<Rightarrow> True)
           then fst v else TAtom Bot)"

fun \<Gamma>::"ex_term \<Rightarrow> ex_type" where
  "\<Gamma> (Var v) = \<Gamma>\<^sub>v v"
| "\<Gamma> (Fun (attack _) _) = TAtom Attack"
| "\<Gamma> (Fun (pubkey _) _) = TAtom Value"
| "\<Gamma> (Fun encodingsecret _) = TAtom PrivFunSec"
| "\<Gamma> (Fun (pubconst \<tau> _) _) = TAtom \<tau>"
| "\<Gamma> (Fun f T) = TComp f (map \<Gamma> T)"


subsection {* Locale interpretation: typed model *}
lemma assm7: "arity c = 0 \<Longrightarrow> \<exists>a. \<forall>X. \<Gamma> (Fun c X) = TAtom a" by (cases c) simp_all

lemma assm8: "0 < arity f \<Longrightarrow> \<Gamma> (Fun f X) = TComp f (map \<Gamma> X)" by (cases f) simp_all

lemma assm9: "infinite {c. \<Gamma> (Fun c []) = TAtom a \<and> public c}"
proof -
  let ?T = "(range (pubconst a))::ex_fun set"
  have *:
      "\<And>x y::nat. x \<in> UNIV \<Longrightarrow> y \<in> UNIV \<Longrightarrow> (pubconst a x = pubconst a y) = (x = y)"
      "\<And>x::nat. x \<in> UNIV \<Longrightarrow> pubconst a x \<in> ?T"
      "\<And>y::ex_fun. y \<in> ?T \<Longrightarrow> \<exists>x \<in> UNIV. y = pubconst a x"
    by auto
  have "?T \<subseteq> {c. \<Gamma> (Fun c []) = TAtom a \<and> public c}" by auto
  moreover have "\<exists>f::nat \<Rightarrow> ex_fun. bij_betw f UNIV ?T"
    using bij_betwI'[OF *] by blast
  hence "infinite ?T" by (metis nat_not_finite bij_betw_finite)
  ultimately show ?thesis using infinite_super by blast
qed

lemma assm10: "TComp f T \<sqsubseteq> \<Gamma> t \<Longrightarrow> arity f > 0"
proof (induction rule: \<Gamma>.induct)
  case (1 x)
  hence *: "TComp f T \<sqsubseteq> \<Gamma>\<^sub>v x" by simp
  hence "\<Gamma>\<^sub>v x \<noteq> TAtom Bot" unfolding \<Gamma>\<^sub>v_def by force
  hence "\<forall>t \<in> subterms (fst x). case t of
            (TComp f T) \<Rightarrow> arity f > 0 \<and> arity f = length T
          | _ \<Rightarrow> True"
    unfolding \<Gamma>\<^sub>v_def by argo
  thus ?case using * unfolding \<Gamma>\<^sub>v_def subtermeq_def by fastforce 
qed auto

lemma assm11: "wf\<^sub>t\<^sub>r\<^sub>m (\<Gamma> (Var x))"
proof -
  have "wf\<^sub>t\<^sub>r\<^sub>m (\<Gamma>\<^sub>v x)" unfolding \<Gamma>\<^sub>v_def wf\<^sub>t\<^sub>r\<^sub>m_def by auto 
  thus ?thesis by simp
qed

interpretation typed_model arity public Ana \<Gamma>
by standard (metis assm7, metis assm8, metis assm9, metis assm10, metis assm11, metis assm6)


subsection {* Locale interpretation: stateful typed model *}
lemma assm12: "Ana (Fun f T) = (K,M) \<Longrightarrow> Ana (Fun f T \<cdot> \<delta>) = (K \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>,M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>)"
by (cases f) (auto elim!: Ana\<^sub>c\<^sub>r\<^sub>y\<^sub>p\<^sub>t.elims Ana\<^sub>s\<^sub>i\<^sub>g\<^sub>n.elims)

interpretation stateful_typed_model arity public Ana \<Gamma> pair
by standard (rule arity.simps, rule assm12)

type_synonym ex_stateful_strand_step = "(ex_fun,ex_var) stateful_strand_step"
type_synonym ex_stateful_strand = "(ex_fun,ex_var) stateful_strand"

lemma Ana_const: "arity c = 0 \<Longrightarrow> Ana (Fun c []) = ({},{})"
by (cases c) auto


subsection {* Locale interpretation: labeled stateful typed model *}
interpretation labeled_typed_model arity public Ana \<Gamma> l1 l2
by standard simp

interpretation labeled_stateful_typed_model arity public Ana \<Gamma> pair l1 l2
..

type_synonym ex_labeled_stateful_strand_step =
  "(ex_fun,ex_var,ex_lbl) labeled_stateful_strand_step"

type_synonym ex_labeled_stateful_strand =
  "(ex_fun,ex_var,ex_lbl) labeled_stateful_strand"



section {* Automatically Proving Type-Flaw Resistance of the Keyserver Protocols *}
subsection {* Definitions *}
definition is_subterm_closed where
  "is_subterm_closed M \<equiv> \<Union>(subterms ` set M) \<subseteq> set M"

definition is_Ana_closed where
  "is_Ana_closed M \<equiv> \<forall>s \<in> set M. case Ana s of (K,T) \<Rightarrow> K \<subseteq> set M"

definition is_TComp_var_instance_closed where
  "is_TComp_var_instance_closed M \<equiv> \<forall>v \<in> fv\<^sub>s\<^sub>e\<^sub>t (set M).
     case \<Gamma>\<^sub>v v of
       (TComp f T) \<Rightarrow> (\<exists>t \<in> set M.
         case t of
           (Fun g U) \<Rightarrow> (\<Gamma> (Fun g U) = \<Gamma>\<^sub>v v \<and> (\<forall>u \<in> set U. is_Var u) \<and> distinct U)
         | _ \<Rightarrow> False)
     | _ \<Rightarrow> True"

definition is_pairwise_var_disjoint where
  "is_pairwise_var_disjoint M \<equiv> \<forall>s \<in> set M. \<forall>t \<in> set M. s \<noteq> t \<longrightarrow> fv s \<inter> fv t = {}"

definition wf\<^sub>t\<^sub>r\<^sub>m_comp::"ex_term \<Rightarrow> bool" where
  "wf\<^sub>t\<^sub>r\<^sub>m_comp t \<equiv> (\<forall>t \<in> subterms t. case t of
                      (Fun f T) \<Rightarrow> arity f = length T
                    | _ \<Rightarrow> True)"

lemma wf\<^sub>t\<^sub>r\<^sub>m_code[code]: "wf\<^sub>t\<^sub>r\<^sub>m t = wf\<^sub>t\<^sub>r\<^sub>m_comp t"
proof
  show "wf\<^sub>t\<^sub>r\<^sub>m_comp t \<Longrightarrow> wf\<^sub>t\<^sub>r\<^sub>m t" unfolding wf\<^sub>t\<^sub>r\<^sub>m_comp_def wf\<^sub>t\<^sub>r\<^sub>m_def subtermeq_def by auto
  
  { fix s assume s: "s \<in> subterms t" "wf\<^sub>t\<^sub>r\<^sub>m t"
    hence "case s of Var x \<Rightarrow> True | Fun f T \<Rightarrow> arity f = length T" 
      unfolding wf\<^sub>t\<^sub>r\<^sub>m_comp_def wf\<^sub>t\<^sub>r\<^sub>m_def subtermeq_def by (cases s) auto
  } 
  thus "wf\<^sub>t\<^sub>r\<^sub>m t \<Longrightarrow> wf\<^sub>t\<^sub>r\<^sub>m_comp t" by (metis wf\<^sub>t\<^sub>r\<^sub>m_comp_def)
qed

definition is_wellformed_terms where
  "is_wellformed_terms M \<equiv> \<forall>t \<in> set M. wf\<^sub>t\<^sub>r\<^sub>m_comp t"

definition "max_var t \<equiv> Max (insert 0 (snd ` fv t))"
abbreviation "max_var' M \<equiv> Max (insert 0 (snd ` fv\<^sub>s\<^sub>e\<^sub>t (set M)))"

definition "var_rename n v \<equiv> Var (fst v, snd v + Suc n)"
definition "var_rename_inv n v \<equiv> Var (fst v, snd v - Suc n)"

fun var_disj' where
  "var_disj' _ [] = []"
| "var_disj' n (t#T) = t \<cdot> var_rename n#var_disj' (max_var t + Suc n) T"

definition "var_disj M \<equiv> var_disj' (max_var' M) M"

fun comp_subterms::"ex_term \<Rightarrow> ex_term list" where
  "comp_subterms (Var x) = [Var x]"
| "comp_subterms (Fun f T) = Fun f T#concat (map comp_subterms T)"

fun comp_trms\<^sub>s\<^sub>s\<^sub>t where
  "comp_trms\<^sub>s\<^sub>s\<^sub>t [] = []"
| "comp_trms\<^sub>s\<^sub>s\<^sub>t (Send t#S) = t#comp_trms\<^sub>s\<^sub>s\<^sub>t S"
| "comp_trms\<^sub>s\<^sub>s\<^sub>t (Receive t#S) = t#comp_trms\<^sub>s\<^sub>s\<^sub>t S"
| "comp_trms\<^sub>s\<^sub>s\<^sub>t (Equality t t'#S) = t#t'#comp_trms\<^sub>s\<^sub>s\<^sub>t S"
| "comp_trms\<^sub>s\<^sub>s\<^sub>t (Inequality _ t t'#S) = t#t'#comp_trms\<^sub>s\<^sub>s\<^sub>t S"
| "comp_trms\<^sub>s\<^sub>s\<^sub>t (Insert t s#S) = t#s#comp_trms\<^sub>s\<^sub>s\<^sub>t S"
| "comp_trms\<^sub>s\<^sub>s\<^sub>t (Delete t s#S) = t#s#comp_trms\<^sub>s\<^sub>s\<^sub>t S"
| "comp_trms\<^sub>s\<^sub>s\<^sub>t (InSet t s#S) = t#s#comp_trms\<^sub>s\<^sub>s\<^sub>t S"
| "comp_trms\<^sub>s\<^sub>s\<^sub>t (NotInSet _ t s#S) = t#s#comp_trms\<^sub>s\<^sub>s\<^sub>t S"

fun comp_setops\<^sub>s\<^sub>s\<^sub>t where
  "comp_setops\<^sub>s\<^sub>s\<^sub>t [] = []"
| "comp_setops\<^sub>s\<^sub>s\<^sub>t (Insert t s#S) = to_pair (t,s)#comp_setops\<^sub>s\<^sub>s\<^sub>t S"
| "comp_setops\<^sub>s\<^sub>s\<^sub>t (Delete t s#S) = to_pair (t,s)#comp_setops\<^sub>s\<^sub>s\<^sub>t S"
| "comp_setops\<^sub>s\<^sub>s\<^sub>t (InSet t s#S) = to_pair (t,s)#comp_setops\<^sub>s\<^sub>s\<^sub>t S"
| "comp_setops\<^sub>s\<^sub>s\<^sub>t (NotInSet _ t s#S) = to_pair (t,s)#comp_setops\<^sub>s\<^sub>s\<^sub>t S"
| "comp_setops\<^sub>s\<^sub>s\<^sub>t (_#S) = comp_setops\<^sub>s\<^sub>s\<^sub>t S"

fun comp_tfr\<^sub>s\<^sub>s\<^sub>t\<^sub>p where
  "comp_tfr\<^sub>s\<^sub>s\<^sub>t\<^sub>p (Equality t t') = (mgu t t' \<noteq> None \<longrightarrow> \<Gamma> t = \<Gamma> t')"
| "comp_tfr\<^sub>s\<^sub>s\<^sub>t\<^sub>p (Inequality X t t') = (
    (\<forall>x \<in> (fv t \<union> fv t')-set X. \<exists>a. \<Gamma> (Var x) = TAtom a) \<or>
    (\<forall>u \<in> subterms t \<union> subterms t'.
      case u of
        Fun f T \<Rightarrow> T = [] \<or> (\<exists>s \<in> set T. s \<notin> Var ` set X)
      | _ \<Rightarrow> True
    ))"
| "comp_tfr\<^sub>s\<^sub>s\<^sub>t\<^sub>p (NotInSet X t t') = (
    (\<forall>u \<in> subterms (to_pair (t,t')).
      case u of
        Fun f T \<Rightarrow> T = [] \<or> (\<exists>s \<in> set T. s \<notin> Var ` set X)
      | _ \<Rightarrow> True
    ))"
| "comp_tfr\<^sub>s\<^sub>s\<^sub>t\<^sub>p _ = True"

definition comp_tfr\<^sub>s\<^sub>e\<^sub>t where
  "comp_tfr\<^sub>s\<^sub>e\<^sub>t M \<equiv>
    is_wellformed_terms M \<and>
    is_subterm_closed M \<and>
    is_Ana_closed M \<and>
    is_TComp_var_instance_closed M \<and>
    is_pairwise_var_disjoint (var_disj M) \<and>
    (\<forall>s \<in> set (var_disj M). \<forall>t \<in> set (var_disj M).
      is_Fun s \<and> is_Fun t \<and> \<Gamma> s \<noteq> \<Gamma> t \<longrightarrow> mgu s t = None)"

definition comp_tfr where
  "comp_tfr S M \<equiv>
    list_all comp_tfr\<^sub>s\<^sub>s\<^sub>t\<^sub>p S \<and>
    set (comp_trms\<^sub>s\<^sub>s\<^sub>t S) \<union> set (comp_setops\<^sub>s\<^sub>s\<^sub>t S) \<subseteq> set M \<and>
    comp_tfr\<^sub>s\<^sub>e\<^sub>t M"


subsection {* Lemma: comp_tfr_set implies tfr_set *}
lemma var_rename_inv_comp: "t \<cdot> (var_rename n \<circ>\<^sub>s var_rename_inv n) = t"
proof (induction t)
  case (Fun f T)
  hence "map (\<lambda>t. t \<cdot> var_rename n \<circ>\<^sub>s var_rename_inv n) T = T" by (simp add: map_idI) 
  thus ?case by (metis subst_apply_term.simps(2)) 
qed (simp add: var_rename_def var_rename_inv_def)

lemma var_rename_fv_disjoint:
  "fv s \<inter> fv (t \<cdot> var_rename (max_var s)) = {}"
proof -
  have 1: "\<forall>v \<in> fv s. snd v \<le> max_var s" unfolding max_var_def by simp
  have 2: "\<And>n. \<forall>v \<in> fv (t \<cdot> var_rename n). snd v > n" unfolding var_rename_def by (induct t) auto
  show ?thesis using 1 2 by force
qed

lemma var_rename_is_renaming:
  "subst_range (var_rename n) \<subseteq> range Var"
  "subst_range (var_rename_inv n) \<subseteq> range Var"
unfolding var_rename_def var_rename_inv_def by auto

lemma var_rename_wt: "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t (var_rename n)" "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t (var_rename_inv n)"
by (auto simp add: var_rename_def var_rename_inv_def wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t_def \<Gamma>\<^sub>v_def)

lemma var_disj_is_var_renaming:
  assumes "t \<in> set M"
  shows "\<exists>m \<in> set (var_disj M). \<exists>\<delta>. wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta> \<and> subst_range \<delta> \<subseteq> range Var \<and> t = m \<cdot> \<delta>"
  using assms
proof -
  { fix n
    have "\<exists>m \<in> set (var_disj' n M). \<exists>\<delta>. wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta> \<and> subst_range \<delta> \<subseteq> range Var \<and> t = m \<cdot> \<delta>"
      using assms
    proof (induction M arbitrary: n)
      case (Cons m M)
      show ?case
      proof (cases "t \<in> set M")
        case False
        hence "t = m" using Cons.prems by simp
        hence "t = m \<cdot> var_rename n \<cdot> var_rename_inv n" by (metis var_rename_inv_comp subst_comp)
        moreover have "m \<cdot> var_rename n \<in> set (var_disj' n (m#M))" by simp
        ultimately show ?thesis by (metis var_rename_is_renaming(2) var_rename_wt(2))
      qed (simp add: Cons.IH del: subst_range.simps)
    qed simp
  } thus ?thesis by (metis var_disj_def)
qed

lemma var_disj_is_var_renaming':
  assumes "m \<in> set M" "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta>" "s = m \<cdot> \<delta>" "is_Fun m"
  shows "\<exists>m' \<in> set (var_disj M). \<exists>\<theta>. wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<theta> \<and> s = m' \<cdot> \<theta> \<and> is_Fun m'"
proof -
  obtain ma da where ma: "ma\<in>set (var_disj M)" "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t da" "subst_range da \<subseteq> range Var" "m = ma \<cdot> da"
    using var_disj_is_var_renaming[OF assms(1)] by moura
  have "is_Fun ma" using ma(3,4) assms(4) by fastforce
  thus ?thesis using assms(2,3) ma by (metis (full_types) subst_subst_compose wt_subst_compose) 
qed

lemma trms\<^sub>s\<^sub>s\<^sub>t_is_comp_trms\<^sub>s\<^sub>s\<^sub>t: "trms\<^sub>s\<^sub>s\<^sub>t A = set (comp_trms\<^sub>s\<^sub>s\<^sub>t A)"
proof (induction A)
  case (Cons a A) thus ?case by (cases a) auto
qed simp

lemma setops\<^sub>s\<^sub>s\<^sub>t_is_comp_setops\<^sub>s\<^sub>s\<^sub>t: "setops\<^sub>s\<^sub>s\<^sub>t A = set (comp_setops\<^sub>s\<^sub>s\<^sub>t A)"
proof (induction A)
  case (Cons a A) thus ?case by (cases a) auto
qed simp

lemma tfr\<^sub>s\<^sub>s\<^sub>t\<^sub>p_is_comp_tfr\<^sub>s\<^sub>s\<^sub>t\<^sub>p: "tfr\<^sub>s\<^sub>s\<^sub>t\<^sub>p a = comp_tfr\<^sub>s\<^sub>s\<^sub>t\<^sub>p a"
proof (cases a)
  case (Equality t t')
  thus ?thesis
    using mgu_always_unifies[of t _ t'] mgu_gives_MGU[of t t']
    by auto
qed (auto split: "term.splits")

lemma is_TComp_var_instance_closedD:
  assumes x: "x \<in> fv\<^sub>s\<^sub>e\<^sub>t (set M)" "\<Gamma>\<^sub>v x = TComp f T"
    and closed: "is_TComp_var_instance_closed M"
  shows "\<exists>g U. Fun g U \<in> set M \<and> \<Gamma> (Fun g U) = \<Gamma>\<^sub>v x \<and> (\<forall>u \<in> set U. is_Var u) \<and> distinct U"
proof -
  let ?P = "\<lambda>g U v. \<Gamma> (Fun g U) = \<Gamma>\<^sub>v v \<and> (\<forall>u \<in> set U. is_Var u) \<and> distinct U"
  let ?Q = "\<lambda>v. (\<exists>t \<in> set M. case t of (Fun g U) \<Rightarrow> ?P g U v | _ \<Rightarrow> False)"
  let ?R = "\<lambda>v. case \<Gamma>\<^sub>v v of (TComp f T) \<Rightarrow> ?Q v | _ \<Rightarrow> True"
  have "?R x" using x(1) closed unfolding is_TComp_var_instance_closed_def by blast
  hence "?Q x" using x(2) by simp
  then obtain t where t: "t \<in> set M" "case t of Var va \<Rightarrow> False | Fun g U \<Rightarrow> ?P g U x" by moura
  then obtain g U where "t = Fun g U" by (metis (no_types, lifting) is_FunE term.case_eq_if) 
  thus ?thesis using t by auto
qed

lemma is_TComp_var_instance_closedD':
  assumes "x \<in> fv\<^sub>s\<^sub>e\<^sub>t (set M)" "TComp f T \<sqsubseteq> \<Gamma>\<^sub>v x"
    and closed: "is_TComp_var_instance_closed M" "is_subterm_closed M"
    and wf: "is_wellformed_terms M"
  shows "\<exists>g U. Fun g U \<in> set M \<and> \<Gamma> (Fun g U) = TComp f T \<and> (\<forall>u \<in> set U. is_Var u) \<and> distinct U"
using assms(1,2)
proof (induction "\<Gamma>\<^sub>v x" arbitrary: x)
  case (Fun g U)
  note IH = Fun.hyps(1)
  have g: "arity g > 0" "public g" using Fun.hyps(2) fun_type_inv[of "Var x"] by simp_all
  then obtain V where V:
      "Fun g V \<in> set M" "\<Gamma> (Fun g V) = \<Gamma>\<^sub>v x" "\<forall>v \<in> set V. \<exists>x. v = Var x"
      "distinct V" "length U = length V"
    using is_TComp_var_instance_closedD[OF Fun.prems(1) Fun.hyps(2)[symmetric] closed(1)]
    by (metis Fun.hyps(2) fun_type_id_eq fun_type_length_eq is_VarE)
  hence U: "U = map \<Gamma> V" using fun_type[OF g(1)] Fun.hyps(2) by simp
  hence 1: "\<And>v. v \<in> set V \<Longrightarrow> \<Gamma> v \<in> set U" by simp
  have 2: "\<And>v. v \<in> set V \<Longrightarrow> v \<in> set M"
    using closed(2) V(1) unfolding is_subterm_closed_def
    by force
  hence 3: "\<And>y. Var y \<in> set V \<Longrightarrow> y \<in> fv\<^sub>s\<^sub>e\<^sub>t (set M)"
    using fv_subset[OF 2] by force
  show ?case
  proof (cases "TComp f T = \<Gamma>\<^sub>v x")
    case True thus ?thesis using V by fastforce
  next
    case False
    then obtain u where u: "u \<in> set U" "TComp f T \<sqsubseteq> u"
      using Fun.prems(2) Fun.hyps(2) unfolding subterm_def by moura
    then obtain y where y: "Var y \<in> set V" "\<Gamma>\<^sub>v y = u" using U V(3) by auto
    show ?thesis using IH[OF _ 3[OF y(1)]] u y(2) by metis
  qed
qed simp

lemma TComp_var_instance_wt_subst_exists:
  assumes gT: "\<Gamma> (Fun g T) = TComp g (map \<Gamma> U)" "wf\<^sub>t\<^sub>r\<^sub>m (Fun g T)"
    and U: "\<forall>u \<in> set U. \<exists>y. u = Var y" "distinct U"
  shows "\<exists>\<theta>. wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<theta> \<and> wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range \<theta>) \<and> Fun g T = Fun g U \<cdot> \<theta>"
proof -
  define the_i where the_i: "the_i \<equiv> \<lambda>y. THE x. x < length U \<and> U ! x = Var y"
  define \<theta> where \<theta>: "\<theta> \<equiv> \<lambda>y. if Var y \<in> set U then T ! the_i y else Var y"

  have g: "arity g > 0" using gT(1,2) fun_type_inv(1) by blast

  have UT: "length U = length T" using fun_type_length_eq gT(1) by auto

  have 1: "\<And>y. Var y \<in> set U \<Longrightarrow> the_i y < length U \<and> U ! the_i y = Var y"
    using theI'[OF distinct_Ex1[OF U(2)]] the_i by simp

  have 2: "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<theta>"
    using \<theta> 1 gT(1) fun_type[OF g] UT
    unfolding wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t_def
    by (metis (no_types, lifting) nth_map term.inject(2))

  have "\<forall>i<length T. U ! i \<cdot> \<theta> = T ! i"
    using \<theta> 1 U(1) UT distinct_Ex1[OF U(2)] in_set_conv_nth
    by (metis (no_types, lifting) subst_apply_term.simps(1))
  hence "T = map (\<lambda>t. t \<cdot> \<theta>) U" by (simp add: UT nth_equalityI)
  hence 3: "Fun g T = Fun g U \<cdot> \<theta>" by simp

  have "subst_range \<theta> \<subseteq> set T" using \<theta> 1 U(1) UT by (auto simp add: subst_domain_def)
  hence 4: "wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range \<theta>)" using gT(2) wf_trm_param by auto

  show ?thesis by (metis 2 3 4)
qed

lemma TComp_var_instance_closed_has_Fun:
  assumes "x \<in> fv\<^sub>s\<^sub>e\<^sub>t (set M)" "t \<sqsubseteq> \<delta> x" "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta>" "\<Gamma>\<^sub>v x = TComp f T" "\<Gamma> t = TComp g T'"
    and t_fun: "is_Fun t"
    and closed: "is_TComp_var_instance_closed M" "is_subterm_closed M"
    and wf: "wf\<^sub>t\<^sub>r\<^sub>m (\<delta> x)" "is_wellformed_terms M"
  shows "\<exists>m \<in> set M. \<exists>\<theta>. wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<theta> \<and> wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range \<theta>) \<and> t = m \<cdot> \<theta> \<and> is_Fun m"
proof -
  obtain T'' where T'': "t = Fun g T''" using assms(5) t_fun fun_type_id_eq by blast
  have g: "arity g > 0" "public g" using assms(5) fun_type_inv[of t] by simp_all
  have "TComp g T' \<sqsubseteq> \<Gamma>\<^sub>v x" using assms(2,3,5)
    by (metis \<Gamma>.simps(1) wf(1) subtermeq_imp_subtermtypeeq wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t_def) 
  then obtain U where U:
      "Fun g U \<in> set M" "\<Gamma> (Fun g U) = TComp g T'" "\<forall>u\<in>set U. \<exists>y. u = Var y"
      "distinct U" "length T'' = length U"
    using is_TComp_var_instance_closedD'[OF assms(1) _ closed wf(2)]
    by (metis assms(5) T'' fun_type_id_eq fun_type_length_eq is_VarE)
  hence UT': "T' = map \<Gamma> U" using fun_type[OF g(1)] assms(5) by simp

  show ?thesis
    using TComp_var_instance_wt_subst_exists UT' T'' U(1,3,4) assms(2,5) wf(1) wf_trm_subtermeq
    by (metis term.disc(2))
qed

lemma comp_tfr\<^sub>s\<^sub>e\<^sub>t_SMP_D_aux:
  assumes hyps: "t' \<sqsubset> t" and wf: "wf\<^sub>t\<^sub>r\<^sub>m t'" and prems: "is_Fun t'"
    and IH: "(\<exists>f. t = Fun f []) \<or>
             (\<exists>m\<in>set M. \<exists>\<delta>. wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta> \<and> wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range \<delta>) \<and> t = m \<cdot> \<delta> \<and> is_Fun m)"
    and closed: "is_subterm_closed M" "is_TComp_var_instance_closed M"
    and wf_M: "is_wellformed_terms M"
  shows "(\<exists>f. t' = Fun f []) \<or>
         (\<exists>m\<in>set M. \<exists>\<delta>. wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta> \<and> wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range \<delta>) \<and> t' = m \<cdot> \<delta> \<and> is_Fun m)"
using hyps
proof (cases "\<exists>f. t = Fun f [] \<or> t' = Fun f []")
  case False
  note F = False
  then obtain m \<delta> where m:
      "m \<in> set M" "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta>" "t = m \<cdot> \<delta>" "is_Fun m" "wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range \<delta>)"
    using IH by moura
  obtain f T where fT: "t' = Fun f T" "arity f > 0"
    using wf F prems
    by (metis is_FunE length_greater_0_conv subtermeqI' wf\<^sub>t\<^sub>r\<^sub>m_def)
  show ?thesis
  proof (cases "\<exists>x \<in> fv m. t' \<sqsubseteq> \<delta> x")
    case True
    then obtain x where x: "x \<in> fv m" "t' \<sqsubseteq> \<delta> x" by moura
    have 1: "x \<in> fv\<^sub>s\<^sub>e\<^sub>t (set M)" using m(1) x(1) by auto
    have 2: "is_Fun (\<delta> x)" using prems x(2) by auto
    have 3: "wf\<^sub>t\<^sub>r\<^sub>m (\<delta> x)" using m(5) by (simp add: wf_trm_img_subst)
    have "\<not>(\<exists>f. \<delta> x = Fun f [])" using F x(2) by auto
    hence "\<exists>f T. \<Gamma>\<^sub>v x = TComp f T" using 2 3 m(2)
      by (metis (no_types) fun_type is_FunE length_greater_0_conv subtermeqI'
                           \<Gamma>.simps(1) wf\<^sub>t\<^sub>r\<^sub>m_def wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t_def)
    moreover have "\<exists>f T. \<Gamma> t' = Fun f T"
      using False prems wf
      by (metis (no_types) fun_type is_FunE length_greater_0_conv subtermeqI' wf\<^sub>t\<^sub>r\<^sub>m_def)
    ultimately show ?thesis
      using TComp_var_instance_closed_has_Fun 1 x(2) m(2) prems closed 3 wf_M
      by metis
  next
    case False
    then obtain m' where m': "m' \<sqsubseteq> m" "t' = m' \<cdot> \<delta>" "is_Fun m'"
      using hyps m(3) subterm_subst_not_img_subterm
      unfolding subterm_def by blast
    hence "m' \<in> set M" using closed(1) m(1) unfolding is_subterm_closed_def by auto
    thus ?thesis using m'(2,3) m(2,4,5) fT by auto
  qed
qed auto

lemma comp_tfr\<^sub>s\<^sub>e\<^sub>t_SMP_D_aux':
  assumes "t \<in> SMP (set M)"
    and closed: "is_subterm_closed M" "is_Ana_closed M" "is_TComp_var_instance_closed M"
    and wf_M: "is_wellformed_terms M"
  shows "\<forall>x \<in> fv t. \<exists>y \<in> fv\<^sub>s\<^sub>e\<^sub>t (set M). \<Gamma> (Var y) = \<Gamma> (Var x)"
using assms(1)
proof (induction t rule: SMP.induct)
  case (MP t)
  thus ?case
  proof (induction t)
    case (Var x)
    hence "x \<in> fv\<^sub>s\<^sub>e\<^sub>t (set M)" using fv_subset[OF MP.hyps] by force
    thus ?case using Var by auto
  next
    case (Fun f T)
    have "\<And>s. s \<in> set T \<Longrightarrow> s \<in> set M"
      using closed(1) Fun.prems
      unfolding is_subterm_closed_def
      by force
    thus ?case using Fun.IH by auto
  qed
next
  case (Subterm t t')
  hence "fv t' \<subseteq> fv t" using subtermeq_vars_subset unfolding subterm_def by auto
  thus ?case using Subterm.IH by auto
next
  case (Substitution t \<delta>)
  note IH = Substitution.IH
  show ?case
  proof
    fix x assume x: "x \<in> fv (t \<cdot> \<delta>)"
    then obtain y where y: "y \<in> fv t" "\<Gamma> (Var x) \<sqsubseteq> \<Gamma> (Var y)"
      using Substitution.hyps(2,3)
      by (metis subst_apply_img_var subtermeqI' subtermeq_imp_subtermtypeeq
                vars_iff_subtermeq wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t_def)
    let ?P = "\<lambda>x. \<exists>y \<in> fv\<^sub>s\<^sub>e\<^sub>t (set M). \<Gamma> (Var y) = \<Gamma> (Var x)"
    show "?P x" using y IH
    proof (induction "\<Gamma> (Var y)" arbitrary: y t)
      case (Var a)
      hence "\<Gamma> (Var x) = \<Gamma> (Var y)" by auto
      thus ?case using Var(2,4) by auto
    next
      case (Fun f T)
      obtain z where z: "z \<in> fv\<^sub>s\<^sub>e\<^sub>t (set M)" "\<Gamma> (Var z) = \<Gamma> (Var y)" using Fun.prems(1,3) by blast
      show ?case
      proof (cases "\<Gamma> (Var x) = \<Gamma> (Var y)")
        case True thus ?thesis using Fun.prems by auto
      next
        case False
        then obtain \<tau> where \<tau>: "\<tau> \<in> set T" "\<Gamma> (Var x) \<sqsubseteq> \<tau>" using Fun.prems(2) Fun.hyps(2) by auto
        then obtain U where U:
            "Fun f U \<in> set M" "\<Gamma> (Fun f U) = \<Gamma> (Var z)" "\<forall>u \<in> set U. \<exists>v. u = Var v" "distinct U"
          using is_TComp_var_instance_closedD'[OF z(1) _ closed(3,1) wf_M] Fun.hyps(2) z(2)
          by (metis \<Gamma>.simps(1) fun_type_id_eq subtermeqI' is_VarE)
        hence 1: "\<forall>x \<in> fv (Fun f U). \<exists>y\<in>fv\<^sub>s\<^sub>e\<^sub>t (set M). \<Gamma> (Var y) = \<Gamma> (Var x)"
          using closed(1) U(1) unfolding is_subterm_closed_def by force
        have "arity f > 0" using U(2) z(2) Fun.hyps(2) fun_type_inv(1) by auto 
        hence "\<Gamma> (Fun f U) = TComp f (map \<Gamma> U)" using fun_type by auto
        then obtain u where u: "Var u \<in> set U" "\<Gamma> (Var u) = \<tau>"
          using \<tau>(1) U(2,3) z(2) Fun.hyps(2) by auto
        show ?thesis
          using Fun.hyps(1)[of u "Fun f U"] u \<tau> 1
          by force
      qed
    qed
  qed
next
  case (Ana t K T k)
  have "fv k \<subseteq> fv t" using Ana_keys_fv[OF Ana.hyps(2)] Ana.hyps(3) by auto
  thus ?case using Ana.IH by auto
qed

lemma comp_tfr\<^sub>s\<^sub>e\<^sub>t_SMP_D:
  assumes "t \<in> SMP (set M)" "is_Fun t"
    and wf_M: "is_wellformed_terms M"
    and closed: "is_subterm_closed M" "is_Ana_closed M" "is_TComp_var_instance_closed M"
  shows "(\<exists>f. t = Fun f []) \<or>
         (\<exists>m \<in> set M. \<exists>\<delta>. wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta> \<and> wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range \<delta>) \<and> t = m \<cdot> \<delta> \<and> is_Fun m)"
using assms(1,2)
proof (induction t rule: SMP.induct)
  case (MP t)
  moreover have "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t Var" "wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range Var)" "t = t \<cdot> Var" by simp_all
  ultimately show ?case by blast 
next
  case (Subterm t t')
  hence t_fun: "is_Fun t" by auto
  have wf: "wf\<^sub>t\<^sub>r\<^sub>m t'"
    using Subterm.hyps(1,2) wf_M(1) SMP_wf_trm wf\<^sub>t\<^sub>r\<^sub>m_code
    unfolding is_wellformed_terms_def
    by (metis wf_trm_subterm)
  note * = Subterm.hyps(2) wf Subterm.prems Subterm.IH[OF t_fun] closed(1,3) wf_M
  show ?case by (rule comp_tfr\<^sub>s\<^sub>e\<^sub>t_SMP_D_aux[OF *])
next
  case (Substitution t \<delta>)
  have wf: "wf\<^sub>t\<^sub>r\<^sub>m t"
    using Substitution.hyps(1) wf_M(1) SMP_wf_trm wf\<^sub>t\<^sub>r\<^sub>m_code
    unfolding is_wellformed_terms_def
    by metis
  hence wf': "wf\<^sub>t\<^sub>r\<^sub>m (t \<cdot> \<delta>)" using Substitution.hyps(3) wf_trm_subst by blast
  have wf_img: "wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range \<delta>)" using Substitution.hyps(3) wf_trms_subst_img_eq by blast
  show ?case
  proof (cases "\<Gamma> t")
    case (Var a)
    hence 1: "\<Gamma> (t \<cdot> \<delta>) = TAtom a" using Substitution.hyps(2) by (metis wt_subst_trm'') 
    show ?thesis using TAtom_term_cases[OF wf' 1] is_FunE[OF Substitution.prems] by auto
  next
    case (Fun f T)
    hence 1: "\<Gamma> (t \<cdot> \<delta>) = TComp f T" using Substitution.hyps(2) by (metis wt_subst_trm'')
    have 2: "\<not>(\<exists>f. t = Fun f [])" using Fun TComp_term_cases[OF wf] by auto
    obtain T'' where T'': "t \<cdot> \<delta> = Fun f T''"
      using 1 2 fun_type_id_eq Fun Substitution.prems
      by auto
    have f: "arity f > 0" "public f" using fun_type_inv[OF 1] by metis+

    show ?thesis
    proof (cases t)
      case (Fun g U)
      then obtain m \<theta> where m: "m \<in> set M" "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<theta>" "t = m \<cdot> \<theta>" "is_Fun m" "wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range \<theta>)"
        using Substitution.IH Fun 2 by moura
      have "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t (\<theta> \<circ>\<^sub>s \<delta>)" "t \<cdot> \<delta> = m \<cdot> (\<theta> \<circ>\<^sub>s \<delta>)" "wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range (\<theta> \<circ>\<^sub>s \<delta>))"
        using wt_subst_compose[OF m(2) Substitution.hyps(2)] m(3)
              wf_trms_subst_compose[OF m(5) wf_img]
        by auto
      thus ?thesis using m(1,4) by metis
    next
      case (Var x)
      then obtain y where y: "y \<in> fv\<^sub>s\<^sub>e\<^sub>t (set M)" "\<Gamma> (Var y) = \<Gamma> (Var x)"
        using comp_tfr\<^sub>s\<^sub>e\<^sub>t_SMP_D_aux'[OF Substitution.hyps(1) closed wf_M] Fun
        by moura
      hence 3: "\<Gamma>\<^sub>v y = TComp f T" using Var Fun by simp
      then obtain h V where V:
          "Fun h V \<in> set M" "\<Gamma> (Fun h V) = \<Gamma>\<^sub>v y" "\<forall>u \<in> set V. \<exists>z. u = Var z" "distinct V"
        by (metis is_VarE is_TComp_var_instance_closedD[OF y(1) 3 closed(3)])
      moreover have "length T'' = length V" using 3 V(2) fun_type_length_eq 1 T'' by auto 
      ultimately have TV: "T = map \<Gamma> V"
        by (metis fun_type[OF f(1)] 3 fun_type_id_eq term.inject(2))


      obtain \<theta> where \<theta>: "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<theta>" "wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range \<theta>)" "t \<cdot> \<delta> = Fun h V \<cdot> \<theta>"
        using TComp_var_instance_wt_subst_exists 1 3 T'' TV V(2,3,4) wf'
        by (metis fun_type_id_eq)

      have 9: "\<Gamma> (Fun h V) = \<Gamma> (\<delta> x)" using y(2) Substitution.hyps(2) V(2) 1 3 Var by auto

      show ?thesis using Var \<theta> 9 V(1) by force
    qed
  qed
next
  case (Ana t K T k)
  have 1: "is_Fun t" using Ana.hyps(2,3) by auto
  then obtain f U where U: "t = Fun f U" by moura

  have 2: "fv k \<subseteq> fv t" using Ana_keys_fv[OF Ana.hyps(2)] Ana.hyps(3) by auto

  have wf_t: "wf\<^sub>t\<^sub>r\<^sub>m t"
    using SMP_wf_trm[OF Ana.hyps(1)] wf\<^sub>t\<^sub>r\<^sub>m_code wf_M
    unfolding is_wellformed_terms_def
    by auto
  hence wf_k: "wf\<^sub>t\<^sub>r\<^sub>m k"
    using Ana_keys_wf'[OF Ana.hyps(2)] wf\<^sub>t\<^sub>r\<^sub>m_code Ana.hyps(3)
    by auto

  have "U \<noteq> []"
    using Ana_const Ana.hyps(2,3) U wf_t 1 TComp_term_cases[OF wf_t]
    by (metis empty_iff fun_type is_Var_def neq0_conv prod.inject term.inject(2))
  then obtain m \<delta> where m: "m \<in> set M" "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta>" "wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range \<delta>)" "t = m \<cdot> \<delta>" "is_Fun m"
    using Ana.IH[OF 1] U by auto
  hence "Ana (t \<cdot> \<delta>) = (K \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>,T \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>)" using Ana_subst' U Ana.hyps(2) by auto
  obtain Km Tm where Ana_m: "Ana m = (Km,Tm)" by moura
  hence "Ana (m \<cdot> \<delta>) = (Km \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>,Tm \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>)"
    using Ana_subst' U m(4) is_FunE[OF m(5)] Ana.hyps(2)
    by metis
  then obtain km where km: "km \<in> Km" "k = km \<cdot> \<delta>" using Ana.hyps(2,3) m(4) by auto
  hence km_in: "km \<in> set M" using closed(2) Ana_m m(1) unfolding is_Ana_closed_def by force
  thus ?case using m(2,3) km(2)
  proof (cases "is_Fun km")
    case False
    note F = False
    then obtain x where x: "km = Var x" by auto
    hence 3: "x \<in> fv\<^sub>s\<^sub>e\<^sub>t (set M)" using fv_subset[OF km_in] by auto
    obtain kf kT where kf: "k = Fun kf kT" using Ana.prems by auto
    show ?thesis
    proof (cases "\<exists>f. k = Fun f []")
      case False
      hence 4: "arity kf > 0" using wf_k kf TAtom_term_cases const_type by fastforce
      then obtain kT' where kT': "\<Gamma> k = TComp kf kT'" by (simp add: fun_type kf) 
      then obtain V where V:
          "Fun kf V \<in> set M" "\<Gamma> (Fun kf V) = \<Gamma>\<^sub>v x" "\<forall>u \<in> set V. \<exists>v. u = Var v"
          "distinct V" "is_Fun (Fun kf V)"
        using x m(2) km(2) is_TComp_var_instance_closedD[OF 3 _ closed(3)]
        by (metis fun_type_id_eq term.disc(2) \<Gamma>.simps(1) wt_subst_trm'' is_VarE)
      have 5: "kT' = map \<Gamma> V"
        using fun_type[OF 4] x kT' km(2) m(2) V(2)
        by (metis \<Gamma>.simps(1) term.inject(2) wt_subst_trm'')
      thus ?thesis
        using TComp_var_instance_wt_subst_exists wf_k kf 4 V(3,4) kT' V(1,5)
        by metis
    qed simp
  qed auto
qed

lemma tfr\<^sub>s\<^sub>e\<^sub>t_if_comp_tfr\<^sub>s\<^sub>e\<^sub>t:
  assumes "comp_tfr\<^sub>s\<^sub>e\<^sub>t M"
  shows "tfr\<^sub>s\<^sub>e\<^sub>t (set M)"
proof -
  have *: "is_wellformed_terms M"
          "is_subterm_closed M" "is_Ana_closed M"
          "is_TComp_var_instance_closed M"
          "is_pairwise_var_disjoint (var_disj M)"
    using assms by (metis comp_tfr\<^sub>s\<^sub>e\<^sub>t_def)+

  { fix s t \<delta> assume st: "s \<in> SMP (set M) - Var`\<V>" "t \<in> SMP (set M) - Var`\<V>" "\<Gamma> s \<noteq> \<Gamma> t"
    hence **: "s \<in> SMP (set M)" "is_Fun s" "t \<in> SMP (set M)" "is_Fun t" by auto
    have "\<not>(\<exists>\<delta>. Unifier \<delta> s t)"
    proof (cases "\<exists>f. s = Fun f [] \<or> t = Fun f []")
      case True thus ?thesis using st(3) **(2,4) by auto
    next
      case False
      then obtain \<sigma> s0 \<theta> t0 where
            m: "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<sigma>" "s0 \<in> set (var_disj M)" "s = s0 \<cdot> \<sigma>" "is_Fun s0"
        and n: "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<theta>" "t0 \<in> set (var_disj M)" "t = t0 \<cdot> \<theta>" "is_Fun t0"
        using comp_tfr\<^sub>s\<^sub>e\<^sub>t_SMP_D[OF **(1,2) *(1,2,3,4)]
              comp_tfr\<^sub>s\<^sub>e\<^sub>t_SMP_D[OF **(3,4) *(1,2,3,4)]
              var_disj_is_var_renaming'
        by moura
      hence "\<Gamma> s = \<Gamma> s0" "\<Gamma> t = \<Gamma> t0" by (metis wt_subst_trm'')+
      hence "\<Gamma> s0 \<noteq> \<Gamma> t0" by (metis st(3))
      hence "\<not>(\<exists>\<delta>. Unifier \<delta> s0 t0)"
        by (metis assms m(2,4) n(2,4) comp_tfr\<^sub>s\<^sub>e\<^sub>t_def mgu_None_is_subst_neq)
      thus ?thesis
        using vars_term_disjoint_imp_unifier *(5) m(2,3) n(2,3)
        by (metis (no_types, hide_lams) is_pairwise_var_disjoint_def subst_comp)
    qed
  } thus ?thesis by (metis tfr\<^sub>s\<^sub>e\<^sub>t_def)
qed

lemma tfr\<^sub>s\<^sub>s\<^sub>t_if_comp_tfr:
  assumes "comp_tfr S M"
  shows "tfr\<^sub>s\<^sub>s\<^sub>t S"
unfolding tfr\<^sub>s\<^sub>s\<^sub>t_def
proof
  show "tfr\<^sub>s\<^sub>e\<^sub>t (trms\<^sub>s\<^sub>s\<^sub>t S \<union> setops\<^sub>s\<^sub>s\<^sub>t S)"
    by (metis trms\<^sub>s\<^sub>s\<^sub>t_is_comp_trms\<^sub>s\<^sub>s\<^sub>t setops\<^sub>s\<^sub>s\<^sub>t_is_comp_setops\<^sub>s\<^sub>s\<^sub>t
              tfr_subset tfr\<^sub>s\<^sub>e\<^sub>t_if_comp_tfr\<^sub>s\<^sub>e\<^sub>t assms comp_tfr_def)

  have "list_all comp_tfr\<^sub>s\<^sub>s\<^sub>t\<^sub>p S" by (metis assms comp_tfr_def)
  thus "list_all tfr\<^sub>s\<^sub>s\<^sub>t\<^sub>p S" by (induct S) (simp_all add: tfr\<^sub>s\<^sub>s\<^sub>t\<^sub>p_is_comp_tfr\<^sub>s\<^sub>s\<^sub>t\<^sub>p)
qed

subsection {* Theorem: Type-flaw resistance of the keyserver example from the CSF18 paper *}
abbreviation "PK n \<equiv> Var (TAtom Value,n)"
abbreviation "A n \<equiv> Var (TAtom Agent,n)"

abbreviation "ringset t \<equiv> Fun ring [Fun encodingsecret [], t]"
abbreviation "validset t t' \<equiv> Fun valid [Fun encodingsecret [], t, t']"
abbreviation "revokedset t t' \<equiv> Fun revoked [Fun encodingsecret [], t, t']"
abbreviation "eventsset \<equiv> Fun events [Fun encodingsecret []]"

(* Note: We will use S\<^sub>k\<^sub>s as a constraint, but it actually represents all steps that might occur
         in the protocol *)
abbreviation S\<^sub>k\<^sub>s::"(ex_fun,ex_var) stateful_strand_step list" where
  "S\<^sub>k\<^sub>s \<equiv> [
    Insert (Fun (attack 0) []) eventsset,
    Delete (PK 0) (validset (A 0) (A 0)),
    NotInSet [(TAtom Agent,0)] (PK 0) (revokedset (A 0) (A 0)),
    NotInSet [(TAtom Agent,0)] (PK 0) (validset (A 0) (A 0)),
    Insert (PK 0) (validset (A 0) (A 0)),
    Insert (PK 0) (ringset (A 0)),
    Insert (PK 0) (revokedset (A 0) (A 0)),
    InSet (PK 0) (validset (A 0) (A 0)),
    InSet (PK 0) (ringset (A 0)),
    Receive (Fun invkey [Fun encodingsecret [], PK 0]),
    Receive (Fun sign [Fun invkey [Fun encodingsecret [], PK 0], Fun pair' [A 0, PK 0]]),
    Send (Fun invkey [Fun encodingsecret [], PK 0]),
    Send (Fun sign [Fun invkey [Fun encodingsecret [], PK 0], Fun pair' [A 0, PK 0]])
]"

theorem "tfr\<^sub>s\<^sub>s\<^sub>t S\<^sub>k\<^sub>s"
proof -
  have "comp_tfr S\<^sub>k\<^sub>s (concat (map comp_subterms (comp_trms\<^sub>s\<^sub>s\<^sub>t S\<^sub>k\<^sub>s@comp_setops\<^sub>s\<^sub>s\<^sub>t S\<^sub>k\<^sub>s)))" by eval
  thus ?thesis by (rule tfr\<^sub>s\<^sub>s\<^sub>t_if_comp_tfr)
qed


subsection {* Theorem: Type-flaw resistance of the keyserver examples from the ESORICS18 paper *}
abbreviation "signmsg t t' \<equiv> Fun sign [t, t']"
abbreviation "cryptmsg t t' \<equiv> Fun crypt [t, t']"
abbreviation "invkeymsg t \<equiv> Fun invkey [Fun encodingsecret [], t]"
abbreviation "updatemsg a b c d \<equiv> Fun update [a,b,c,d]"
abbreviation "pwmsg t t' \<equiv> Fun pw [t, t']"

abbreviation "beginauthset n t t' \<equiv> Fun (beginauth n) [Fun encodingsecret [], t, t']"
abbreviation "endauthset n t t' \<equiv> Fun (endauth n) [Fun encodingsecret [], t, t']"
abbreviation "pubkeysset t \<equiv> Fun pubkeys [Fun encodingsecret [], t]"
abbreviation "seenset t \<equiv> Fun seen [Fun encodingsecret [], t]"

(* Note: S'\<^sub>k\<^sub>s contains the (slightly over-approximated) steps that can occur in the
         reachable constraints of \<P>\<^sub>k\<^sub>s,\<^sub>1 and \<P>\<^sub>k\<^sub>s,\<^sub>2 modulo variable renaming *)
abbreviation S'\<^sub>k\<^sub>s::"ex_labeled_stateful_strand_step list" where
  "S'\<^sub>k\<^sub>s \<equiv> [
(* constraint steps from the first protocol (duplicate steps are ignored) *)

    (* rule R^1_1 *)
    (ln l1, Send (invkeymsg (PK 0))),
    (\<star>,     InSet (PK 0) (validset (A 0) (A 1))),
    (ln l1, Receive (Fun (attack 0) [])),

    (* rule R^2_1 *)
    (ln l1, Send (signmsg (invkeymsg (PK 0)) (Fun pair' [A 0, PK 0]))),
    (\<star>,     InSet (PK 0) (validset (A 0) (A 1))),
    (\<star>,     NotInSet [(TAtom Agent,0), (TAtom Agent,1)] (PK 0) (validset (A 0) (A 1))),
    (ln l1, NotInSet [(TAtom Agent,0), (TAtom Agent,1)] (PK 0) (revokedset (A 0) (A 1))),
    (\<star>,     NotInSet [] (PK 0) (beginauthset 0 (A 0) (A 1))),

    (* rule R^3_1 *)
    (\<star>,     InSet (PK 0) (beginauthset 0 (A 0) (A 1))),
    (\<star>,     InSet (PK 0) (endauthset 0 (A 0) (A 1))),

    (* rule R^4_1 *)
    (\<star>,     Receive (PK 0)),
    (\<star>,     Receive (invkeymsg (PK 0))),

    (* rule R^5_1 *)
    (ln l1, Insert (PK 0) (ringset (A 0))),
    (\<star>,     Insert (PK 0) (validset (A 0) (A 1))),
    (\<star>,     Insert (PK 0) (beginauthset 0 (A 0) (A 1))),
    (\<star>,     Insert (PK 0) (endauthset 0 (A 0) (A 1))),

    (* rule R^6_1 *)
    (ln l1, InSet (PK 0) (ringset (A 0))),
    (ln l1, Delete (PK 0) (ringset (A 0))),
    
    (* rule R^7_1 *)
    (\<star>,     NotInSet [] (PK 0) (endauthset 0 (A 0) (A 1))),
    (\<star>,     Delete (PK 0) (validset (A 0) (A 1))),
    (ln l1, Insert (PK 0) (revokedset (A 0) (A 1))),

    (* rule R^8_1 *)
    (* nothing new *)

    (* rule R^9_1 *)
    (ln l1, Send (PK 0)),
    
    (* rule R^10_1 *)
    (ln l1, Send (Fun (attack 0) [])),


(* constraint steps from the second protocol (duplicate steps are ignored) *)
    (* rule R^2_1 *)
    (ln l2, Send (invkeymsg (PK 0))),
    (\<star>,     InSet (PK 0) (validset (A 0) (A 1))),
    (ln l2, Receive (Fun (attack 1) [])),

    (* rule R^2_2 *)
    (ln l2, Send (cryptmsg (PK 0) (updatemsg (A 0) (A 1) (PK 1) (pwmsg (A 0) (A 1))))),
    (ln l2, InSet (PK 0) (pubkeysset (A 0))),
    (ln l2, NotInSet [(TAtom Agent,0)] (PK 0) (pubkeysset (A 0))),
    (ln l2, NotInSet [(TAtom Agent,0)] (PK 0) (seenset (A 0))),

    (* rule R^3_2 *)
    (\<star>,     InSet (PK 0) (beginauthset 1 (A 0) (A 1))),
    (\<star>,     InSet (PK 0) (endauthset 1 (A 0) (A 1))),

    (* rule R^4_2 *)
    (\<star>,     Receive (PK 0)),
    (\<star>,     Receive (invkeymsg (PK 0))),

    (* rule R^5_2 *)
    (ln l2, InSet (PK 0) (pubkeysset (A 0))),
    (\<star>,     Insert (PK 0) (beginauthset 1 (A 0) (A 1))),
    (ln l2, Receive (cryptmsg (PK 0) (updatemsg (A 0) (A 1) (PK 1) (pwmsg (A 0) (A 1))))),

    (* rule R^6_2 *)
    (\<star>,     NotInSet [] (PK 0) (endauthset 1 (A 0) (A 1))),
    (\<star>,     Insert (PK 0) (validset (A 0) (A 1))),
    (\<star>,     Insert (PK 0) (endauthset 1 (A 0) (A 1))),
    (ln l2, Insert (PK 0) (seenset (A 0))),

    (* rule R^7_2 *)
    (ln l2, Receive (pwmsg (A 0) (A 1))),

    (* rule R^8_2 *)
    (* nothing new *)

    (* rule R^9_2 *)
    (ln l2, Insert (PK 0) (pubkeysset (A 0))),

    (* rule R^10_2 *)
    (ln l2, Send (Fun (attack 1) []))
]"

theorem "tfr\<^sub>s\<^sub>s\<^sub>t (unlabel S'\<^sub>k\<^sub>s)"
proof -
  let ?S = "unlabel S'\<^sub>k\<^sub>s"
  have "comp_tfr ?S (concat (map comp_subterms (comp_trms\<^sub>s\<^sub>s\<^sub>t ?S@comp_setops\<^sub>s\<^sub>s\<^sub>t ?S)))" by eval
  thus ?thesis by (rule tfr\<^sub>s\<^sub>s\<^sub>t_if_comp_tfr)
qed

end
